@@ -1,5 +1,6 @@
 /* OpenCP Module Player
- * copyright (c) '94-'10 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
+ * copyright (c) 1994-'10 Niklas Beisert <nbeisert@physik.tu-muenchen.de>
+ * copyright (c) 2004-'22 Stian Skjelstad <stian.skjelstad@gmail.com>
  *
  * OGGPlay - Player for Ogg Vorbis files
  *
@@ -34,15 +35,19 @@
 #include <vorbis/codec.h>
 #include <vorbis/vorbisfile.h>
 #include "types.h"
+#include "cpiface/gif.h"
+#include "cpiface/jpeg.h"
+#include "cpiface/png.h"
+#include "dev/deviplay.h"
+#include "dev/mcp.h"
+#include "dev/player.h"
+#include "dev/plrasm.h"
+#include "dev/ringbuffer.h"
+#include "filesel/filesystem.h"
+#include "oggplay.h"
 #include "stuff/imsrtns.h"
 #include "stuff/timer.h"
 #include "stuff/poll.h"
-#include "dev/player.h"
-#include "dev/deviplay.h"
-#include "dev/plrasm.h"
-#include "dev/ringbuffer.h"
-
-#include "oggplay.h"
 
 static int current_section;
 
@@ -52,18 +57,21 @@ static int signedout;
 static uint32_t samprate;
 static uint8_t reversestereo;
 
-/* static unsigned long amplify; TODO */
-static unsigned long voll,volr;
+static uint32_t voll,volr;
+static int vol;
+static int bal;
 static int pan;
 static int srnd;
 
+static int opt25_50;
+static char opt25[26];
+static char opt50[51];
+
 static int16_t *buf16=NULL;
 static uint32_t bufpos;
 static uint32_t buflen;
 static void *plrbuf;
 
-/*static uint32_t amplify;*/
-
 static OggVorbis_File ov;
 static int oggstereo;
 static int oggrate;
@@ -83,6 +91,11 @@ static int inpause;
 
 static volatile int clipbusy=0;
 
+static struct ocpfilehandle_t *oggfile;
+
+static int (*_GET)(int ch, int opt);
+static void (*_SET)(int ch, int opt, int val);
+
 #define PANPROC \
 do { \
 	float _rs = rs, _ls = ls; \
@@ -207,7 +220,6 @@ void __attribute__ ((visibility ("internal"))) oggIdle(void)
 		return;
 	}
 
-
 	{
 		uint32_t bufplayed;
 
@@ -591,36 +603,510 @@ void __attribute__ ((visibility ("internal"))) oggIdle(void)
 	clipbusy--;
 }
 
-void __attribute__ ((visibility ("internal"))) oggSetAmplify(uint32_t amp)
+static size_t read_func (void *ptr, size_t size, size_t nmemb, void *token)
 {
-/*
-	amplify=amp;
-	float v[9];
-	float ampf=(float)vols[9]*(amplify/65536.0)/65536.0;
-	int i;
-	for (i=0; i<9; i++)
-		v[i]=ampf*vols[i];
-	rawogg.ioctl(ampegdecoder::ioctlsetstereo, v, 4*9);
-	*/
+	uint64_t retval;
+	retval = oggfile->read (oggfile, ptr, size * nmemb);
+	return retval / size;
 }
 
-static int close_func(void *datasource)
+static int seek_func (void *token, ogg_int64_t offset, int whence)
+{
+	switch (whence)
+	{
+		case SEEK_SET:
+			if (oggfile->seek_set (oggfile, offset) < 0)
+			{
+				return -1;
+			}
+			break;
+		case SEEK_END:
+			if (oggfile->seek_end (oggfile, offset) < 0)
+			{
+				return -1;
+			}
+			break;
+		case SEEK_CUR:
+			if (oggfile->seek_cur (oggfile, offset) < 0)
+			{
+				return -1;
+			}
+			break;
+		default:
+			return -1;
+	}
+	return oggfile->getpos (oggfile);
+}
+
+static int close_func(void *token)
 {
 	return 0;
 }
 
-int __attribute__ ((visibility ("internal"))) oggOpenPlayer(FILE *oggf)
+static long tell_func (void *token)
 {
+	return oggfile->getpos (oggfile);
+}
+
+
+struct ogg_comment_t __attribute__ ((visibility ("internal"))) **ogg_comments;
+int                  __attribute__ ((visibility ("internal")))   ogg_comments_count;
+struct ogg_picture_t __attribute__ ((visibility ("internal")))  *ogg_pictures;
+int                  __attribute__ ((visibility ("internal")))   ogg_pictures_count;
+
+static void add_comment2(const char *title, const char *value)
+{
+	int n = 0;
+	for (n = 0; n < ogg_comments_count; n++)
+	{
+		int res = strcmp (ogg_comments[n]->title, title);
+		if (res == 0)
+		{
+			// append to at this point
+			ogg_comments[n] = realloc (ogg_comments[n], sizeof (*ogg_comments[n]) + sizeof (ogg_comments[n]->value[0]) * (ogg_comments[n]->value_count + 1));
+			ogg_comments[n]->value[ogg_comments[n]->value_count++] = strdup(value);
+			return;
+		}
+		if (res < 0)
+		{
+			continue;
+		} else {
+			// insert it at this point
+			goto insert;
+		}
+	}
+
+insert:
+	ogg_comments = realloc (ogg_comments, sizeof (ogg_comments[0]) * (ogg_comments_count+1));
+	memmove (ogg_comments + n + 1, ogg_comments + n, (ogg_comments_count - n) * sizeof (ogg_comments[0]));
+	ogg_comments[n] = malloc (sizeof (*ogg_comments[n]) + sizeof (ogg_comments[n]->value[0]));
+	ogg_comments[n]->title = strdup (title);
+	ogg_comments[n]->value_count = 1;
+	ogg_comments[n]->value[0] = strdup (value);
+	ogg_comments_count++;
+}
+
+static void add_picture(const uint16_t actual_width,
+		        const uint16_t actual_height,
+			uint8_t *data_bgra,
+			const char *description,
+			const uint32_t description_length,
+			const uint32_t picture_type)
+{
+	ogg_pictures = realloc (ogg_pictures, sizeof (ogg_pictures[0]) * (ogg_pictures_count + 1));
+	ogg_pictures[ogg_pictures_count].picture_type = picture_type;
+	ogg_pictures[ogg_pictures_count].description = malloc (description_length + 1);
+	memcpy (ogg_pictures[ogg_pictures_count].description, description, description_length);
+	ogg_pictures[ogg_pictures_count].description[description_length] = 0;
+	ogg_pictures[ogg_pictures_count].width = actual_width;
+	ogg_pictures[ogg_pictures_count].height = actual_height;
+	ogg_pictures[ogg_pictures_count].data_bgra = data_bgra;
+	ogg_pictures[ogg_pictures_count].scaled_width = 0;
+	ogg_pictures[ogg_pictures_count].scaled_height = 0;
+	ogg_pictures[ogg_pictures_count].scaled_data_bgra = 0;
+
+	ogg_pictures_count++;
+}
+
+#ifdef __GNUC__
+#pragma GCC diagnostic push
+#pragma GCC diagnostic ignored "-Wunused-but-set-variable"
+#endif
+static void add_picture_binary(const uint8_t *src, unsigned int srclen)
+{
+	uint32_t picture_type;
+	uint32_t mime_length;
+	const uint8_t *mime;
+	uint32_t description_length;
+	const uint8_t *description;
+	uint32_t width;
+	uint32_t height;
+	uint32_t bpp;
+	uint32_t colors; // if GIF color palette
+	uint32_t data_length;
+	const uint8_t *data;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	picture_type = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	mime_length = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < mime_length)
+	{
+		return;
+	}
+	mime = src;
+	src += mime_length; srclen -= mime_length;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	description_length = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < description_length)
+	{
+		return;
+	}
+	description = src;
+	src += description_length; srclen -= description_length;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	width = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	height = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	bpp = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	colors = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < 4)
+	{
+		return;
+	}
+	data_length = (src[0]<<24) | (src[1]<<16) | (src[2]<<8) | src[3];
+	src += 4; srclen -= 4;
+
+	if (srclen < data_length)
+	{
+		return;
+	}
+	data = src;
+	src += data_length; srclen -= data_length;
+
+#if 0
+	if ((mime_length == 3) && (!strncasecmp (mime, "-->", 3)))
+	{
+		// TODO - URL
+	}
+#endif
+#ifdef HAVE_LZW
+	if ((mime_length == 9) && (!strncasecmp ((const char *)mime, "image/gif", 9)))
+	{
+		uint16_t actual_height, actual_width;
+		uint8_t *data_bgra;
+		if (!GIF87_try_open_bgra (&actual_width, &actual_height, &data_bgra, data, data_length))
+		{
+			add_picture (actual_width, actual_height, data_bgra, (const char *)description, description_length, picture_type);
+		}
+		return;
+	}
+#endif
+
+	if ((mime_length == 9) && (!strncasecmp ((const char *)mime, "image/png", 9)))
+	{
+		uint16_t actual_height, actual_width;
+		uint8_t *data_bgra;
+		if (!try_open_png (&actual_width, &actual_height, &data_bgra, (uint8_t *)data, data_length))
+		{
+			add_picture (actual_width, actual_height, data_bgra, (const char *)description, description_length, picture_type);
+		}
+		return;
+	}
+
+	if (((mime_length == 9) && (!strncasecmp ((const char *)mime, "image/jpg", 9))) || ((mime_length == 10) && (!strncasecmp ((const char *)mime, "image/jpeg", 10))))
+	{
+		uint16_t actual_height, actual_width;
+		uint8_t *data_bgra;
+		if (!try_open_jpeg (&actual_width, &actual_height, &data_bgra, data, data_length))
+		{
+			add_picture (actual_width, actual_height, data_bgra, (const char *)description, description_length, picture_type);
+		}
+		return;
+	}
+}
+#ifdef __GNUC__
+#pragma GCC diagnostic pop
+#endif
+
+static uint8_t base64_index (const char src)
+{
+	if ((src >= 'A') && (src <= 'Z')) return src - 'A';
+	if ((src >= 'a') && (src <= 'z')) return src - 'a' + 26;
+	if ((src >= '0') && (src <= '9')) return src - '0' + 52;
+	if (src == '+') return 62;
+	if (src == '/') return 63;
+	if (src == '=') return 64;
+	return 65;
+}
+
+static void add_picture_base64(const char *src)
+{
+	int srclen = strlen (src);
+	uint8_t *dst;
+	int dstlen = 0;
+
+	if (srclen < 2)
+	{
+		return;
+	}
+
+	dst = malloc (srclen * 3 / 4); // this is a good estimate, but it might overshoot due to padding
+
+	while (srclen >= 2)
+	{
+		uint8_t tmp1, tmp2, tmp3, tmp4;
+
+		tmp1 = base64_index (*src);
+		src++; srclen--;
+		if (tmp1 >= 64)
+		{
+			break;
+		}
+
+		tmp2 = base64_index (*src);
+		src++; srclen--;
+		if (tmp2 >= 64)
+		{/* this is invalid encoding */
+			break;
+		}
+
+		*dst = (tmp1<<2) | (tmp2>>4); //6 + 2
+		dst++; dstlen++;
+
+		if (!srclen)
+		{
+			break;
+		}
+		tmp3 = base64_index (*src);
+		src++; srclen--;
+		if (tmp3 >= 64)
+		{
+			break;
+		}
+
+		*dst = (tmp2<<4) | (tmp3>>2); //4 + 4
+		dst++; dstlen++;
+
+		if (!srclen)
+		{
+			break;
+		}
+		tmp4 = base64_index (*src);
+		src++; srclen--;
+		if (tmp4 >= 64)
+		{
+			break;
+		}
+
+		*dst = (tmp3<<6) | tmp4; // 2 + 6
+		dst++; dstlen++;
+	}
+
+	dst -= dstlen;
+
+	add_picture_binary(dst, dstlen);
+
+	free (dst);
+}
+
+static void add_comment(const char *src)
+{
+	char *equal, *tmp, *tmp2;
+	if (!strncasecmp (src, "METADATA_BLOCK_PICTURE=", 23))
+	{
+		add_picture_base64(src + 23);
+		return;
+	}
+	equal = strchr (src, '=');
+
+	if (!equal)
+	{
+		return;
+	}
+	if (equal == src)
+	{
+		return;
+	}
+
+	tmp = malloc (equal - src + 1);
+	strncpy (tmp, src, equal - src);
+	tmp[equal-src] = 0;
+
+	if ((tmp[0] >= 'a') && (tmp[0] <= 'z')) tmp[0] -= 0x20;
+
+	for (tmp2 = tmp + 1; *tmp2; tmp2++)
+	{
+		if ((tmp2[0] >= 'A') && (tmp2[0] <= 'Z')) tmp2[0] += 0x20;
+	}
+
+	add_comment2(tmp, src + (equal - src) + 1);
+
+	free (tmp);
+}
+
+char __attribute__ ((visibility ("internal"))) oggLooped(void)
+{
+	return looped == 3;
+}
+
+void __attribute__ ((visibility ("internal"))) oggSetLoop(uint8_t s)
+{
+	donotloop=!s;
+}
+
+void __attribute__ ((visibility ("internal"))) oggPause(uint8_t p)
+{
+	inpause=p;
+}
+
+static void oggSetSpeed (uint16_t sp)
+{
+	if (sp<32)
+		sp=32;
+	oggbufrate=imuldiv(256*sp, oggrate, samprate);
+}
+
+static void oggSetVolume (void)
+{
+	volr=voll=vol*4;
+	if (bal<0)
+		voll=(voll*(64+bal))>>6;
+	else
+		volr=(volr*(64-bal))>>6;
+}
+
+static void SET(int ch, int opt, int val)
+{
+	switch (opt)
+	{
+		case mcpMasterSpeed:
+			oggSetSpeed(val);
+			break;
+		case mcpMasterPitch:
+			break;
+		case mcpMasterSurround:
+			srnd=val;
+			break;
+		case mcpMasterPanning:
+			pan=val;
+			if (reversestereo)
+			{
+				pan = -pan;
+			}
+			oggSetVolume();
+			break;
+		case mcpMasterVolume:
+			vol=val;
+			oggSetVolume();
+			break;
+		case mcpMasterBalance:
+			bal=val;
+			oggSetVolume();
+			break;
+	}
+}
+static int GET(int ch, int opt)
+{
+	return 0;
+}
+
+ogg_int64_t __attribute__ ((visibility ("internal"))) oggGetPos(void)
+{
+#warning TODO, deduct devp buffer!
+	return (oggpos+ogglen-ringbuffer_get_tail_available_samples(oggbufpos))%ogglen;
+}
+
+void __attribute__ ((visibility ("internal"))) oggGetInfo(struct ogginfo *info)
+{
+	static int lastsafe=0;
+	info->pos=oggGetPos();
+	info->len=ogglen;
+	info->rate=oggrate;
+	info->stereo=oggstereo;
+	info->bit16=1;
+	if ((info->bitrate=ov_bitrate_instant (&ov))<0)
+		info->bitrate=lastsafe;
+	else
+		lastsafe=info->bitrate;
+	if (!opt25_50)
+	{
+		vorbis_info *vi = ov_info (&ov, -1);
+		if (vi)
+		{
+			snprintf (opt25, sizeof (opt25), "Ogg Vorbis version %d", vi->version);
+			snprintf (opt50, sizeof (opt50), "Ogg Vorbis version %d, %d channels", vi->version, vi->channels);
+			opt25_50 = 1;
+		}
+	}
+	info->opt25=opt25;
+	info->opt50=opt50;
+}
+
+void __attribute__ ((visibility ("internal"))) oggSetPos(ogg_int64_t pos)
+{
+	pos=(pos+ogglen)%ogglen;
+
+	oggneedseek=1;
+	oggpos=pos;
+	ringbuffer_reset(oggbufpos);
+}
+
+static ov_callbacks callbacks =
+{
+	read_func,
+	seek_func,
+	close_func,
+	tell_func
+};
+int __attribute__ ((visibility ("internal"))) oggOpenPlayer(struct ocpfilehandle_t *oggf)
+{
+	int result;
 	struct vorbis_info *vi;
 
 	if (!plrPlay)
 		return 0;
 
-	fseek(oggf, 0, SEEK_SET);
-	if(ov_open(oggf, &ov, NULL, -1) < 0)
-		return -1; /* we don't bother to do more exact */
-	ov.callbacks.close_func=close_func;
-
+	oggf->seek_set (oggf, 0);
+	if (oggfile)
+	{
+		oggfile->unref (oggfile);
+		oggfile = 0;
+	}
+	oggfile = oggf;
+	oggfile->ref (oggfile);
+	if ((result = ov_open_callbacks(oggfile, &ov, NULL, 0, callbacks)))
+	{
+		switch (result)
+		{
+			case OV_EREAD:      fprintf (stderr, "ov_open_callbacks(): A read from media returned an error.\n"); break;
+			case OV_ENOTVORBIS: fprintf (stderr, "ov_open_callbacks(): Bitstream does not contain any Vorbis data.\n"); break;
+			case OV_EVERSION:   fprintf (stderr, "ov_open_callbacks(): Vorbis version mismatch.\n"); break;
+			case OV_EBADHEADER: fprintf (stderr, "ov_open_callbacks(): Invalid Vorbis bitstream header.\n"); break;
+			case OV_EFAULT:     fprintf (stderr, "ov_open_callbacks(): Internal logic fault; indicates a bug or heap/stack corruption.\n"); break;
+			default:            fprintf (stderr, "ov_open_callbacks(): Unknown error %d\n", result); break;
+		}
+		return 0; /* we don't bother to do more exact */
+	}
 
 	vi=ov_info(&ov,-1);
 	oggstereo=(vi->channels>=2);
@@ -653,7 +1139,21 @@ int __attribute__ ((visibility ("internal"))) oggOpenPlayer(FILE *oggf)
 	current_section=0;
 	oggneedseek=0;
 
-	if (!plrOpenPlayer(&plrbuf, &buflen, plrBufSize * plrRate / 1000))
+	{
+		int i;
+
+		vorbis_comment *vf = 0;
+		vf = ov_comment (&ov, -1);
+		if (vf)
+		{
+			for (i=0; i < vf->comments; i++)
+			{
+				add_comment(vf->user_comments[i]);
+			}
+		}
+	}
+
+	if (!plrOpenPlayer(&plrbuf, &buflen, plrBufSize * plrRate / 1000, oggf))
 	{
 		ringbuffer_free (oggbufpos);
 		oggbufpos = 0;
@@ -665,9 +1165,6 @@ int __attribute__ ((visibility ("internal"))) oggOpenPlayer(FILE *oggf)
 
 	inpause=0;
 	looped=0;
-	oggSetVolume(64, 0, 64, 0);
-/*
-	oggSetAmplify(amplify);   TODO */
 
 	buf16=malloc(sizeof(uint16_t)*buflen*2);
 	if (!buf16)
@@ -698,13 +1195,25 @@ int __attribute__ ((visibility ("internal"))) oggOpenPlayer(FILE *oggf)
 
 		return 0;
 	}
+
+	_SET=mcpSet;
+	_GET=mcpGet;
+	mcpSet=SET;
+	mcpGet=GET;
+
+	mcpNormalize (mcpNormalizeDefaultPlayP);
+
 	active=1;
+	opt25_50 = 0;
+	opt25[0] = 0;
+	opt50[0] = 0;
 
 	return 1;
 }
 
 void __attribute__ ((visibility ("internal"))) oggClosePlayer(void)
 {
+	int i, j;
 	active=0;
 
 	pollClose();
@@ -721,72 +1230,45 @@ void __attribute__ ((visibility ("internal"))) oggClosePlayer(void)
 	buf16=NULL;
 
 	ov_clear(&ov);
-}
 
-char __attribute__ ((visibility ("internal"))) oggLooped(void)
-{
-	return looped == 3;
-}
-
-void __attribute__ ((visibility ("internal"))) oggSetLoop(uint8_t s)
-{
-	donotloop=!s;
-}
-
-void __attribute__ ((visibility ("internal"))) oggPause(uint8_t p)
-{
-	inpause=p;
-}
-
-void __attribute__ ((visibility ("internal"))) oggSetSpeed(uint16_t sp)
-{
-	if (sp<32)
-		sp=32;
-	oggbufrate=imuldiv(256*sp, oggrate, samprate);
-}
-
-void __attribute__ ((visibility ("internal"))) oggSetVolume(uint8_t vol_, int8_t bal_, int8_t pan_, uint8_t opt)
-{
-	pan=pan_;
-	if (reversestereo)
+	for (i=0; i < ogg_comments_count; i++)
 	{
-		pan = -pan;
+		for (j=0; j < ogg_comments[i]->value_count; j++)
+		{
+			free (ogg_comments[i]->value[j]);
+		}
+		free (ogg_comments[i]->title);
+		free (ogg_comments[i]);
 	}
-	volr=voll=vol_*4;
-	if (bal_<0)
-		voll=(voll*(64+bal_))>>6;
-	else
-		volr=(volr*(64-bal_))>>6;
-	srnd=opt;
-}
+	free (ogg_comments);
+	ogg_comments = 0;
+	ogg_comments_count = 0;
 
-ogg_int64_t __attribute__ ((visibility ("internal"))) oggGetPos(void)
-{
-#warning TODO, deduct devp buffer!
-	return (oggpos+ogglen-ringbuffer_get_tail_available_samples(oggbufpos))%ogglen;
-}
+	for (i=0; i < ogg_pictures_count; i++)
+	{
+		free (ogg_pictures[i].data_bgra);
+		free (ogg_pictures[i].scaled_data_bgra);
+		free (ogg_pictures[i].description);
+	}
+	free (ogg_pictures);
+	ogg_pictures = 0;
+	ogg_pictures_count = 0;
 
-void __attribute__ ((visibility ("internal"))) oggGetInfo(struct ogginfo *i)
-{
-	static int lastsafe=0;
-	i->pos=oggGetPos();
-	i->len=ogglen;
-	i->rate=oggrate;
-	i->stereo=oggstereo;
-	i->bit16=1;
-	if ((i->bitrate=ov_bitrate_instant(&ov))<0)
-		i->bitrate=lastsafe;
-	else
-		lastsafe=i->bitrate;
-	i->bitrate/=1000;
-}
+	if (oggfile)
+	{
+		oggfile->unref (oggfile);
+		oggfile = 0;
+	}
 
-void __attribute__ ((visibility ("internal"))) oggSetPos(ogg_int64_t pos)
-{
-	pos=(pos+ogglen)%ogglen;
-
-	oggneedseek=1;
-	oggpos=pos;
-	ringbuffer_reset(oggbufpos);
+	if (_SET)
+	{
+		mcpSet = _SET;
+		_SET = 0;
+	}
+	if (_GET)
+	{
+		mcpGet = _GET;
+		_GET = 0;
+	}
 }
 
